
-- | Let's try some A*-style search algorithm for a fresh change
--
-- We use the "fingertree" library for priority queues: 
-- http://hackage.haskell.org/package/fingertree
--

{-# LANGUAGE PackageImports #-}
module AStar where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Ord

import Control.Monad
import Control.Applicative

import Data.IORef
import System.IO
import System.Timeout

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import qualified "fingertree" Data.PriorityQueue.FingerTree as PQ
import "fingertree" Data.PriorityQueue.FingerTree (PQueue)

import PosSet
import Spatial
import Mine
import Simulation
import SimBetter

--------------------------------------------------------------------------------

simulatePathWithUndoThenRewind :: SimState -> Path -> (a -> IO b) -> (SimState -> IO a) -> IO b
simulatePathWithUndoThenRewind state path cont action = do
  x <- simulatePathWithUndo state path action
  cont x

--------------------------------------------------------------------------------

data AStar = AStar 
  { pathCost      :: SimState -> Path -> Double
  , heuristicDist :: SimState -> Pos -> Pos -> Double 
  }
  
defaultAStar = AStar 
  { pathCost      = \ state path  ->  -(pathScore path)     -- this is *not* admissible, but who cares about theory :)
  , heuristicDist = \ state p q   ->    astarDist p q
  }

-- | It must be unique. Pos more-or-less guarantees it.
-- And sorted by priority. Lexicographic order solves that.
data PrioKey = PrioKey 
  { _priority :: {-# UNPACK #-} !Double  
  , _prioPos  :: {-# UNPACK #-} !Pos
  , _prioPath :: !Path
  }
  deriving (Eq,Ord,Show) -- ,Show)

-- instance Show PrioKey where show (PrioKey prio pos _) = "{PrioKey " ++ show prio ++ " | " ++ show pos ++ "}"

--------------------------------------------------------------------------------

pqToList :: Ord k => PQueue k v -> [k] 
pqToList = go where
  go pq = case PQ.minViewWithKey pq of
    Just ((k,v),rest) -> k : go rest
    Nothing -> []

--------------------------------------------------------------------------------
  
type OpenSet   = PQueue PrioKey ()
type ClosedMap = Map Pos PrioKey

--------------------------------------------------------------------------------

-- | 1/5 second time-out
timeoutAStar :: AStar -> SimState -> Pos -> IO (Maybe Path)
timeoutAStar cfg state tgt = do

  -- timeout corrupts the single state!!!!!!!!!!!!
  frozen <- freeze (_simMine state) :: IO Mine
  clone  <- unsafeThaw frozen
  let tmpstate = state { _simMine = clone }
  
  timeout (200*1000) (fullAStar cfg tmpstate tgt) >>= \m -> return $ case m of
    Just mp -> mp
    Nothing -> Nothing

 
fullAStar :: AStar -> SimState -> Pos -> IO (Maybe Path)
fullAStar cfg state0 tgtpos = 
  do
    go iniOpen iniClosed
    
  where
  
    go :: OpenSet -> ClosedMap -> IO (Maybe Path)
    go open closed = do
      
      --putStrLn "============================"
      --print ("open"  ,pqToList open)
      --print ("closed",closed)
      
      case PQ.minViewWithKey open of
        Nothing -> return Nothing
        Just ( ( pkey@(PrioKey prio ppos path@(Path unpath)) , _) , open1 ) -> do
          -- not really optimal :(
          --simulatePathWithUndo state0 path $ \state -> do
          let continue = return
          simulatePathWithUndoThenRewind state0 path id $ \state -> do
            let pos = _simRobotPos state
            if pos == target_to_reach -- tgtpos 
              then continue $ return (Just path) 
              else do
              
                --putStrLn $ "examining: " ++ show ppos ++ " == " ++ show pos ++ " | " ++ show pkey
              
                -- check if we already visited this place some time in the past (and then are we better now?)
                let b = case Map.lookup pos closed of
                          Just old@(PrioKey oldprio _ _) -> (prio < oldprio)    -- <= hmm hmm volt ellenpelda amikor ez kell
                          Nothing                        -> True 
               
                if not b then continue (go open1 closed) else do 

                  let closed' = Map.insert pos pkey closed 
                                        
                  -- where can we move
                  nexts <- generateMoves meta mine pos
                  
                  --putStrLn $ "potential moves: " ++ show nexts
                  
                  pkeys <- forM nexts $ \move -> do
                    -- let's simulate these steps, if we made the trouble of reconstrucing the state...
                    simulateMoveWithUndo state move $ \state' -> do
                      let pos' = _simRobotPos state'
                          
                          deathpenalty' = case stateEndCond state' of
                            Just Losing      -> -666666
                            Just Inoperative -> -666666
                            _                -> 0
                          justscored' = mmovementScore (_simLastMovement state')
                          score' = pathScore path + justscored' + deathpenalty' + (if (pos==tgtpos) then 100000 else 0)
                                     
                      let step' = Step move score' {-(Status score)-} (_simLastChanges state')
                          path' = Path (step':unpath)
                          fv'   = fvalue state' path' pos'
                      return $ PrioKey fv' pos' path'
                      
                  let open' = foldr (\pkey -> PQ.insert pkey ()) open1 pkeys 
                  continue (go open' closed')

{-                
    insertClosed :: Pos -> PrioKey -> ClosedMap -> ClosedMap
    insertClosed pos pkey cmap = Map.insertWith f pos pkey cmap where
      f new@(PrioKey newf _ _) old@(PrioKey oldf _ _) = if newf < oldf then new else old
-}          

    fvalue state path pos 
      = pathCost      cfg state path 
      + heuristicDist cfg state pos tgtpos  
  
    mine   = _simMine state0
    meta   = _simMeta state0
    srcpos = _simRobotPos state0

    -- teleport atugrik, azt kell csekkolni
    target_to_reach = case Map.lookup tgtpos (_simTrampolins state0) of
      Nothing     -> tgtpos
      Just portal -> portal
      
    inidist   = heuristicDist cfg state0 srcpos tgtpos
    iniOpen   = PQ.singleton (PrioKey inidist srcpos emptyPath) ()   :: OpenSet
    iniClosed = Map.empty                                            :: ClosedMap

--------------------------------------------------------------------------------

dijkstraBest :: Int -> SimState -> IO (Maybe Path)
dijkstraBest radius state = do
  full <- finiteDijkstra radius state
  return $ case sortBy (comparing fst) (Map.elems full) of
    []        -> Nothing
    ((c,p):_) -> case p of { Path [] -> Nothing ; _ -> Just p }
    
-- | We do a Dijkstra in an area of given radius
finiteDijkstra :: Int -> SimState -> IO (Map Pos (Double,Path))
finiteDijkstra radius state0 = 
  do
    alreadySeen <- newIORef (Map.empty)
    wrapper alreadySeen
    readIORef alreadySeen
    
  where
    wrapper alreadySeen = go state0 emptyPath where
    
      go state path@(Path unpath) = do
        let pos = _simRobotPos state
            cost = - (pathScore path)    -- sign!!
            
        seen <- readIORef alreadySeen
        b <- case Map.lookup pos seen of 
          Nothing -> do
            writeIORef alreadySeen $ Map.insert pos (cost,path) seen
            return True
          Just (oldcost,_) -> if (cost < oldcost) 
            then do
              writeIORef alreadySeen $ Map.insert pos (cost,path) seen
              return True
            else return False
            
        when (b && (length unpath < radius)) $ do  
          nexts <- generateMoves meta mine pos       -- where can we move
          
          forM_ nexts $ \next -> do
            simulateMoveWithUndo state next $ \state' -> do
              let deathpenalty' = case stateEndCond state' of
                    Just Losing      -> -666666
                    Just Inoperative -> -666666
                    _                -> 0
                  justscored' = mmovementScore (_simLastMovement state')
                  score' = pathScore path + justscored' + deathpenalty'                           
              let step' = Step next score' (_simLastChanges state')          
              go state' (Path (step':unpath))  
        
    mine   = _simMine state0
    meta   = _simMeta state0
    srcpos = _simRobotPos state0

  
--------------------------------------------------------------------------------
         
main_astar_debug fn x y = do
  (meta,mine) <- loadMap fn 
 
  print ("A* target is: ",(x,y))
  iniState <- initialSimState meta mine
  mpath <- fullAStar defaultAStar iniState (mkPos x y)
  
  case mpath of
    Nothing -> putStrLn "A* failed to find a path"
    Just (Path steps) -> do
      let moves = reverse $ map _stepMove steps
      putStrLn $ encodeMoves moves
      timeMachine meta mine moves

--------------------------------------------------------------------------------
