
{-# LANGUAGE BangPatterns #-}
module Mine where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Maybe

import Control.Monad
import Control.Applicative

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import Data.Char
import System.IO

import Ansi
import PosSet

{-
import Debug.Trace
debug x y = trace ("---" ++ show x ++ "---") y
-}

--------------------------------------------------------------------------------

data Meta = Meta 
  { _metaFlooding    :: Flooding
  , _metaTrampolines :: [Trampoline]
  , _metaHair        :: Hair
  }
  deriving Show

defaultMeta :: Meta
defaultMeta = Meta 
  { _metaFlooding    = defaultFlooding
  , _metaTrampolines = []
  , _metaHair        = defaultHair
  }

insertTrampoline :: Trampoline -> Meta -> Meta
insertTrampoline t (Meta fld trs hair) = Meta fld (t:trs) hair

adjustFlooding :: (Flooding -> Flooding) -> Meta -> Meta
adjustFlooding f (Meta fld trs hair) = Meta (f fld) trs hair

adjustHair :: (Hair -> Hair) -> Meta -> Meta
adjustHair f (Meta fld trs hair) = Meta fld trs (f hair)

--------------------------------------------------------------------------------

data Trampoline = Trampoline 
  { _trampSrc :: (Pos,Char) 
  , _trampTgt :: (Pos,Char) 
  }
  deriving (Eq,Show)
  
findTrampolineFrom :: Meta -> Pos -> Trampoline
findTrampolineFrom (Meta fld trs hair) pos = 
  case find f trs of 
    Just  t -> t
    Nothing -> error ("findTrampolineFrom: " ++ show pos)   
  where 
    f (Trampoline (src,_) (tgt,_)) = src==pos

findTrampolinesTo :: Meta -> Pos -> [Trampoline]
findTrampolinesTo (Meta fld trs hair) pos = filter f trs -- let stuff = filter f trs in debug stuff stuff
  where 
    f (Trampoline (src,_) (tgt,_)) = tgt==pos

--------------------------------------------------------------------------------

data Hair = Hair
  { _growth        :: !Int
  , _initialRazors :: !Int   
  }
  deriving Show
  
defaultHair :: Hair
defaultHair = Hair
  { _growth        = 25
  , _initialRazors = 0 
  }
  
--------------------------------------------------------------------------------

data Flooding = Flooding
  { _waterLevel :: !Int
  , _floodRate  :: !(Maybe Int)
  , _waterProof :: !Int
  }
  deriving (Eq,Show)
  
defaultFlooding :: Flooding
defaultFlooding = Flooding              
  { _waterLevel = 0
  , _floodRate  = Nothing
  , _waterProof = 10
  }

--------------------------------------------------------------------------------
  
parseMetaLine :: String -> Meta -> Meta
parseMetaLine ln = if null ws then id else
  case map toLower (head ws) of
    "water"      -> adjustFlooding $ \old -> old { _waterLevel = read param } 
    "flooding"   -> adjustFlooding $ \old -> old { _floodRate  = rate } where
                      rate = case read param of { 0 -> Nothing ; k -> Just k }
    "waterproof" -> adjustFlooding $ \old -> old { _waterProof = read param }
    "trampoline" -> insertTrampoline $ Trampoline (undefined,head (ws!!1)) (undefined,head (ws!!3)) 
    "growth"     -> adjustHair $ \old -> old { _growth = read param }
    "razors"     -> adjustHair $ \old -> old { _initialRazors = read param }
  where
    ws = words ln
    param = ws!!1

parseMetaLines :: [String] -> Meta
parseMetaLines ls = foldr parseMetaLine defaultMeta ls
   
--------------------------------------------------------------------------------

-- | Left/Right is already used :(
data Dir = U | D | L | R deriving (Eq,Ord,Show)

allDirs = [U,D,L,R]

isHorizontal :: Dir -> Bool
isHorizontal d = case d of { L -> True ; R -> True ; _ -> False }

data Move 
  = Move !Dir 
  | Wait
  | Abort
  | Shave
  deriving (Eq,Ord,Show)  

  
data LiftState = Open | Closed deriving (Eq,Ord,Show)  
  
data Cell
  = Empty 
  | Earth
  | Rock Bool   -- True if higher-order (with hidden lambda), False if normal rock
  | Wall
  | Lambda
  | Lift !LiftState
  | Robot
  | Beard      -- Wadler's beard
  | Razor      -- Hutton's razor
  | TrampSrc !Char         -- ^ trampoline
  | TrampTgt !Char         -- ^ trampoline target
  | Unknown !Char          -- ^ there were some unknown characters in the 1st version of contest10.map 
  | SuperWall              -- ^ the superwall we put around the map
  deriving (Eq,Ord,Show)

type Mine   = Array   Pos Cell
type IOMine = IOArray Pos Cell

cellIsReachable :: Cell -> Bool
cellIsReachable c = case c of
  Empty -> True
  Earth -> True
  Lift Open -> True
  Lambda -> True
  Robot -> True
  Razor -> True
  TrampSrc {} -> True
  _ -> False
  
--------------------------------------------------------------------------------

{- 
-- moved to PosSet.hs

-- We don't use
-- > type Pos = (Int,Int)  
-- since the update order is not the natural order for pairs!
data Pos = Pos {-# UNPACK #-} !Int {-# UNPACK #-} !Int deriving (Eq,Show,Ix)
-- we invert the order, so that there is no need to reverse a list
-- instance Ord Pos where compare (Pos x1 y1) (Pos x2 y2) = compare (y1,x1) (y2,x2)
instance Ord Pos where compare (Pos x1 y1) (Pos x2 y2) = compare (y2,x2) (y1,x1) 

-}

ycoord :: Pos -> Int
ycoord p = snd $ unPos p -- (Pos x y) = y

{-# INLINE manhattanDist #-}
manhattanDist :: Pos -> Pos -> Int
manhattanDist !p1 !p2 {- (Pos x1 y1) (Pos x2 y2) -} = abs (x1-x2) + abs (y1-y2) where
  OldPos !x1 !y1 = unPos' p1 
  OldPos !x2 !y2 = unPos' p2

{-# INLINE manhattanDistf #-}
manhattanDistf :: Pos -> Pos -> Double
manhattanDistf p q = fromIntegral (manhattanDist p q)

{-# INLINE euclideanDist #-}
euclideanDist :: Pos -> Pos -> Double
euclideanDist !p1 !p2 {- (Pos x1 y1) (Pos x2 y2) -} 
  = sqrt $ fromIntegral $ (x1-x2)*(x1-x2) + (x1-x2)*(x1-x2) where
     OldPos !x1 !y1 = unPos' p1 
     OldPos !x2 !y2 = unPos' p2


-- we don't use pure manhattan because of the ambiguity
-- Euclidean will help deciding the direction?
astarDist :: Pos -> Pos -> Double
astarDist p q = 0.6666 * manhattanDistf p q + 0.3333 * euclideanDist p q 

euclideanDistSqr :: Pos -> Pos -> Double
euclideanDistSqr p1 p2 {- (Pos x1 y1) (Pos x2 y2) -}   
 = fromIntegral $ (x1-x2)*(x1-x2) + (x1-x2)*(x1-x2) where
     OldPos !x1 !y1 = unPos' p1 
     OldPos !x2 !y2 = unPos' p2

--------------------------------------------------------------------------------

data EndCond
  = Winning
  | Losing
  | Aborted
  | Inoperative             -- ^ underwater for too long
  deriving (Eq,Ord,Show)

data Result = Result 
  { _endCond   :: EndCond      -- ^ this shouldn't be strict, since we like to put undefined into it :)
  , _lambdaCnt :: Int
  , _moveCnt   :: Int 
  , _history   :: History
  }
  deriving Show

losingCond :: EndCond -> Bool
losingCond Winning = False
losingCond Aborted = False
losingCond _       = True

losingResult :: Result -> Bool
losingResult (Result endcond nlambda nmove hist) = losingCond endcond

resultPoints :: Result -> Int
resultPoints (Result endcond lamcnt movcnt hist) = case endcond of
  Winning     -> 75*lamcnt - movcnt
  Aborted     -> 50*lamcnt - movcnt
  Losing      -> 25*lamcnt - movcnt
  Inoperative -> 25*lamcnt - movcnt

--------------------------------------------------------------------------------
  
newtype History = History [Move] deriving (Eq,Ord)

instance Show History where
  show (History moves) = encodeMoves moves

saveHistory :: FilePath -> History -> IO ()  
saveHistory fpath hist = writeFile fpath (show hist)

readHistory :: FilePath -> IO History
readHistory fpath = do
  text <- readFile fpath
  return $ parseHistory text

parseHistory :: String -> History  
parseHistory = History . parseMoves
   
--------------------------------------------------------------------------------
    
infixl `move`

-- the name move is shadowed in some part of the program...
moveByDir :: Pos -> Dir -> Pos
moveByDir = move

move :: Pos -> Dir -> Pos
move pos d = case d of
  U -> mkPos x (y+1)
  D -> mkPos x (y-1)
  L -> mkPos (x-1) y
  R -> mkPos (x+1) y
  where
    (x,y) = unPos pos
    
-- | Pos is alwayst the /old/ position
data Movement
  = NoMove          !Pos
  | MoveSimple      !Pos !Dir
  | LambdaCollected !Pos !Dir
  | RazorCollected  !Pos !Dir
  | MoveAndPush     !Pos !Dir !Bool    -- bool is rock type flag (HO or normal)    
  | MineCompleted   !Pos !Dir
  | AbortGame       !Pos
  | Trampolined     !Pos !Dir !Pos     -- (old pos, approach, new pos)
  | Shaved          !Pos [Pos]         -- (pos, beards in the neighbourhood)
  | NoRazor         !Pos               -- tried to shave without razor (stupid thing)
  deriving (Eq,Ord,Show)

{-# SPECIALIZE mmovementScore :: Maybe Movement -> Int    #-}
{-# SPECIALIZE mmovementScore :: Maybe Movement -> Double #-}
mmovementScore :: Num a => Maybe Movement -> a
mmovementScore mmm = case mmm of
  Nothing -> -1
  Just mm -> movementScore mm
  
{-# SPECIALIZE movementScore :: Movement -> Int    #-}
{-# SPECIALIZE movementScore :: Movement -> Double #-}
movementScore :: Num a => Movement -> a
movementScore mm = case mm of
  LambdaCollected {}   -> 66  - 1
  RazorCollected  {}   -> 300 - 1
  MineCompleted   {}   -> 1000000 
  MoveAndPush _ _ True -> 15  - 1
  Trampolined     {}   -> 20  - 1
  Shaved _ what        -> 40 * fromIntegral (length what) - 25
  NoRazor {}           -> -1000           -- don't try to shave without razor...
  _                    -> -1  
 
{-
pureCanMove :: Mine -> Pos -> Move -> Maybe Movement
pureCanMove mine pos mv = case mv of
  Abort -> Just $ AbortGame pos
  Wait  -> Just $ NoMove    pos 
  Move dir -> pureCanMove' mine pos dir
  
pureCanMove' :: Mine -> Pos -> Dir -> Maybe Movement
pureCanMove' mine xy dir = 
  case mine ! xy' of
    Empty     -> Just $ MoveSimple      xy dir
    Earth     -> Just $ MoveSimple      xy dir
    Lambda    -> Just $ LambdaCollected xy dir
    Lift Open -> Just $ MineCompleted   xy 
    Rock      -> case mine ! xy'' of
                   Empty -> Just (MoveAndPush xy dir)
                   _     -> Nothing
    _         -> Nothing
  where
    xy'  = xy  `move` dir
    xy'' = xy' `move` dir
-}

--------------------------------------------------------------------------------

-- had to move this to Simulation.hs because of SimState
canMove_noShave :: Meta -> IOMine -> Pos -> Move -> IO (Maybe Movement)
canMove_noShave meta mine pos mv =
  case mv of
    Abort -> return $ Just $ AbortGame pos
    Wait  -> return $ Just $ NoMove    pos 
    Move dir -> canMove' meta mine pos dir
    Shave -> error "canMove_noShave: Shave"
        
canMove' :: Meta -> IOMine -> Pos -> Dir -> IO (Maybe Movement)
canMove' meta mine xy dir = do
  let xy'  = xy  `move` dir
      xy'' = xy' `move` dir
  c' <- readArray mine xy' 
  case c' of
    Empty      -> return $ Just $ MoveSimple      xy dir
    Earth      -> return $ Just $ MoveSimple      xy dir
    Lambda     -> return $ Just $ LambdaCollected xy dir
    Razor      -> return $ Just $ RazorCollected  xy dir
    Lift Open  -> return $ Just $ MineCompleted   xy dir 
    TrampSrc c -> return $ Just $ Trampolined     xy dir zw where
      Trampoline _ (zw,d) = findTrampolineFrom meta xy'
    Rock rflag  -> if isHorizontal dir 
      then do c'' <- readArray mine xy''
              return $ case c'' of
                Empty -> Just (MoveAndPush xy dir rflag)
                _     -> Nothing
      else return Nothing
    _ -> return $ Nothing
  
--------------------------------------------------------------------------------

data Write = Write !Pos !Cell deriving (Eq,Ord,Show)

newtype Undo = Undo Write deriving (Eq,Ord,Show)

data ChangeSet = ChangeSet [Write] [Undo] deriving (Eq,Ord,Show)

unUndo :: Undo -> Write
unUndo (Undo w) = w

executeWrites :: IOMine -> [Write] -> IO ()
executeWrites mine writes =
  forM_ writes $ \(Write pos cell) -> writeArray mine pos cell

executeWritesWithUndo :: IOMine -> [Write] -> IO [Undo]
executeWritesWithUndo mine writes = reverse <$> do         -- we should do undo in the opposite order, /of course/
  forM writes $ \(Write pos cell) -> do
    oldcell <- readArray mine pos
    writeArray mine pos cell
    return $ Undo (Write pos oldcell)
    
executeUndo :: IOMine -> [Undo] -> IO ()
executeUndo mine undos = executeWrites mine (map unUndo undos)
    
isCellWritten :: Pos -> [Write] -> Bool
isCellWritten pos list = case find f list of { Nothing -> False ; Just _ -> True } where
  f (Write p _) = (pos == p)

--------------------------------------------------------------------------------

-- | Computes the writes coming from a realized movement
writeMovement :: Meta -> Movement -> (Pos,[Write])
writeMovement meta movement = do

  case movement of
  
    MoveSimple pos dir -> ( pos' , list ) where 
      list = [ Write pos  Empty
             , Write pos' Robot ] 
      pos'  = move pos  dir 
      
    LambdaCollected pos dir -> ( pos' , list ) where 
      list = [ Write pos  Empty
             , Write pos' Robot ] 
      pos'  = move pos  dir 

    RazorCollected pos dir -> ( pos' , list ) where 
      list = [ Write pos  Empty
             , Write pos' Robot ] 
      pos'  = move pos  dir 

    MoveAndPush pos dir rflag -> ( pos' , list ) where 
      list = [ Write pos   Empty
             , Write pos'  Robot 
             , Write pos'' (Rock rflag)
             ] 
      pos'  = move pos  dir 
      pos'' = move pos' dir 
      
    Shaved pos clear -> ( pos , list ) where
      list = [ Write q Empty | q <- clear ]

    Trampolined pos dir tgtpos -> ( tgtpos , list ) where
      list = (Write pos Empty) : (Write tgtpos Robot) 
           : [ Write src Empty | Trampoline (src,_) _ <- findTrampolinesTo meta tgtpos ]

    MineCompleted pos dir -> ( pos' , list ) where
      list = [ Write pos  Empty
             , Write pos' Robot ] 
      pos'  = move pos  dir 
    
    NoMove        pos -> ( pos , [] )
    NoRazor       pos -> ( pos , [] )
    AbortGame     pos -> ( pos , [] )

--------------------------------------------------------------------------------

type Watch = PSet -- Set Pos

-- | Given a set of writes, we can compute which position to watch in the next step
-- These are the written rocks, and also the cells around a written empty space:
--
-- >   XXX
-- >   X:X
--
-- where : denotes the new empty space
writeWatch :: [Write] -> Watch
writeWatch = foldl' worker setEmpty where
  worker old (Write pos cell) = 
    case cell of
      Rock _    -> setInsert pos old
      Lambda    -> old
      Empty     -> setUnion old (setFromList neighbourhood) 
      Lift Open -> old
      Robot     -> old
      Beard     -> old
      _ -> error $ "writeWatch: unexpected written cell: " ++ show cell
    where
      neighbourhood = 
        [ pos `move` L
        , pos `move` R
        , pos `move` U
        , pos `move` U `move` L
        , pos `move` U `move` R
        ]

-- | Given an initial mine, what to watch.
-- We will be stupid and flag all the rocks
mineWatch :: Mine -> Watch
mineWatch mine = setFromList [ pos | (pos, Rock _) <- assocs mine ]

--------------------------------------------------------------------------------

-- first arg is beard growth trigger flag
updateCell :: Bool -> IOMine -> Pos -> IO [Write]
updateCell = updateCell' False

updateCell' :: Bool -> Bool -> IOMine -> Pos -> IO [Write]
updateCell' beardGrowth noMoreLambdas mine pos = do
  c <- readArr pos
  case c of
    
    Lift _ -> return $ if noMoreLambdas then [ Write pos (Lift Open) ] else []          
    
    Beard -> if not beardGrowth then return [] else do
               wheretogrow <- checkForEmpty mine pos
               return [ Write q Beard | q <- wheretogrow ]
               
    Rock rflag -> do

      let pdown  = pos   `move` D
          pright = pos   `move` R
          pdr    = pdown `move` R
          pleft  = pos   `move` L
          pdl    = pdown `move` L
          
      cdown  <- readArr pdown
      cright <- readArr pright
      cdr    <- readArr pdr
      cleft  <- readArr pleft
      cdl    <- readArr pdl
  
      let myrock = Rock rflag
          handleRock               
            | cdown  == Empty        = [ Write pdown myrock      -- falls down
                                       , Write pos   Empty ]
            
            | cdown  == Empty        = [ Write pdown myrock      -- falls down
                                       , Write pos   Empty ]
            | isRock cdown    && 
              cright == Empty && 
              cdr    == Empty        = [ Write pdr   myrock      -- falls right down
                                       , Write pos   Empty ]
            | isRock cdown &&
              (cright /= Empty || cdr /= Empty) &&
              (cleft  == Empty && cdl == Empty) 
                                     = [ Write pdl   myrock      -- falls left down 
                                       , Write pos   Empty ]
            | cdown  == Lambda && 
              cright == Empty  && 
              cdr    == Empty        = [ Write pdr   myrock      -- falls right on lambda
                                       , Write pos   Empty ]
            | otherwise = []                   
            
      -- Check for shattering
      if not rflag || null handleRock
        then return handleRock 
        else do
          let [ Write pos_r rock, Write pos_e Empty] = handleRock
              below = pos_r `move` D
          cbelow <- readArr below
          if cbelow/=Empty
            then return [ Write pos_r Lambda, Write pos_e Empty]
            else return handleRock    
           
    _ -> return []
    
  where
    readArr :: Pos -> IO Cell
    readArr  = readArray mine

--------------------------------------------------------------------------------

checkLosingCond :: Pos -> [Write] -> IOMine -> IO Bool
checkLosingCond robotpos recentWrites mine = do
  let up = (robotpos `move` U) 
  c <- readArray mine up
  return ((isRock c || c==Lambda) && isCellWritten up recentWrites)
  
--------------------------------------------------------------------------------

parseMove :: Char -> Maybe Move
parseMove c = case toUpper c of
  'U' -> Just $ Move U
  'D' -> Just $ Move D
  'L' -> Just $ Move L
  'R' -> Just $ Move R
  'W' -> Just $ Wait
  'A' -> Just $ Abort
  'S' -> Just $ Shave
  _   -> Nothing

parseMoves :: String -> [Move]
parseMoves = catMaybes . map parseMove
  
encodeDir :: Dir -> Char
encodeDir d = case d of
  U -> 'U'
  D -> 'D'
  L -> 'L'
  R -> 'R'
  
encodeMove :: Move -> Char
encodeMove m = case m of
  Move d -> encodeDir d
  Wait   -> 'W'
  Abort  -> 'A'  
  Shave  -> 'S'

encodeMoves :: [Move] -> [Char]
encodeMoves = map encodeMove
  
--------------------------------------------------------------------------------

parseCell :: Char -> Cell
parseCell c 
  | c>='A' && c<='I'  = TrampSrc c
  | c>='1' && c<='9'  = TrampTgt c
  | otherwise  = case c of
      'R' -> Robot
      '*' -> Rock False
      '@' -> Rock True
      'L' -> Lift Closed
      'O' -> Lift Open
      '.' -> Earth
      '#' -> Wall
      '\\' -> Lambda
      ' ' -> Empty
      'W' -> Beard
      '!' -> Razor
      _   -> Unknown c
  
showCell :: Cell -> Char
showCell c = case c of
  Empty -> ' ' 
  Earth -> '.'
  Rock False -> '*'
  Rock True  -> '@'
  Wall  -> '#'
  Lambda -> '\\'
  Lift Open   -> 'O'
  Lift Closed -> 'L'
  Robot -> 'R'
  TrampSrc c -> c
  TrampTgt c -> c
  Beard -> 'W'
  Razor -> '!'
  Unknown  c -> c
  SuperWall -> ' '

showCellAnsi :: Cell -> String
showCellAnsi c = 
  case c of
    Empty -> " "
    Earth -> "."
    Rock False -> ansi 35 $ "*"
    Rock True  -> ansi 35 $ "@"
    Wall  -> ansi 41 $ ansi 30 $ "#"
    Lambda -> "\\"
    Lift Open   -> "O"
    Lift Closed -> "L"
    Robot -> ansi 37 $ ansi 45 "R"
    TrampSrc c -> ansi 42 $ ansi 37 $ [c]
    TrampTgt c -> ansi 46 $ ansi 37 $ [c]
    Beard -> ansi 43 $ ansi 35 $ "W"
    Razor -> ansi 36 "!"
    Unknown  c -> [c]
    SuperWall  -> " "
  where
    ansi k s = sgr k ++ s ++ sgr 0 -- "\ESC[" ++ show k ++ "m" ++ s ++ "\ESC[0m"

--------------------------------------------------------------------------------

parseMine :: String -> (Meta,Mine)
parseMine text = (meta,mine) where
  (meta1,mine) = parseMine1 text
  meta = meta1 { _metaTrampolines = map worker (_metaTrampolines meta1) }
  worker (Trampoline (_,sc) (_,tc)) = Trampoline (sp,sc) (tp,tc) where
    sp = unsafeFindCell (TrampSrc sc) mine
    tp = unsafeFindCell (TrampTgt tc) mine
  
-- no trampoline position search yet
parseMine1 :: String -> (Meta,Mine)
parseMine1 text = stuff where
  ls = lines text
  stuff = case findIndex isBlank ls of
    Nothing -> ( defaultMeta                , parseMineLines ls          )
    Just k  -> ( parseMetaLines (drop k ls) , parseMineLines (take k ls) )
  isBlank l = null (words l)                           -- ????? maybe not strictly correct ?????
   
parseMineOnly :: String -> Mine
parseMineOnly text = parseMineLines (lines text)

parseMineLines :: [String] -> Mine
parseMineLines ls = mine where
  n = maximum (map length ls) 
  m = length ls
  rows = (map.map) parseCell $ map (\l -> take n (l ++ repeat ' ')) ls
  mine = array (mkPos 1 1, mkPos n m) 
    [ (mkPos x y, cell) | (y,row) <- zip (reverse [1..m]) rows, (x,cell) <- zip [1..n] row ] 

printMine :: Mine -> IO ()
printMine mine = do
  let (OldPos a b, OldPos n m) = oldBounds mine
  forM_ [m,m-1..b] $ \y -> do
    forM_ [a..n] $ \x -> putChar (showCell $ mine!(mkPos x y))
    putStrLn ""

dumpExtendedIOMineAscii :: IOMine -> IO ()
dumpExtendedIOMineAscii mine = do
  (OldPos a b, OldPos n m) <- getOldBounds mine
  forM_ [m-1,m-2..b+1] $ \y -> do
    forM_ [a+1..n-1] $ \x -> do
      c <- readArray mine (mkPos x y)
      putChar (showCell $ c)
    putStrLn ""

printMineWater :: Int -> Mine -> IO ()
printMineWater waterlevel mine = do
  let (OldPos a b, OldPos n m) = oldBounds mine
  forM_ [m,m-1..b] $ \y -> do
    if (y<=waterlevel)
      then do
        withColors (White,Blue) $ do
          -- forM_ [a..n] $ \x -> putChar (showCell $ mine!(Pos x y))
          forM_ [a..n] $ \x -> putStr (showCellAnsi $ mine!(mkPos x y))
        putStrCol Blue " ~~~~~"
      else do 
        forM_ [a..n] $ \x -> putStr (showCellAnsi $ mine!(mkPos x y))
    putStrLn ""

printIOMine :: IOMine -> IO ()
printIOMine mine = freeze mine >>= printMine 

printIOMineWater :: Int -> IOMine -> IO ()
printIOMineWater water mine = freeze mine >>= printMineWater water 

-- | Puts a wall around a mine. Positions remain the same
extendMine :: Mine -> Mine
extendMine old = new where
  (OldPos a b, OldPos n m) = oldBounds old
  new = array (mkPos (a-1) (b-1) , mkPos (n+1) (m+1) ) (walls ++ assocs old)
  walls = [ (xy,SuperWall) | xy<-wallPos ]
  wallPos =  [ mkPos (a-1) y | y<-[b..m] ]
          ++ [ mkPos (n+1) y | y<-[b..m] ]
          ++ [ mkPos x (b-1) | x<-[a-1..n+1] ]
          ++ [ mkPos x (m+1) | x<-[a-1..n+1] ]

extendedMineSize' :: Mine -> (Int,Int)
extendedMineSize' mine = (n-a-1 , m-b-1) where (OldPos a b, OldPos n m) = oldBounds mine

extendedMineSize :: Mine -> Int
extendedMineSize mine = x*y where (x,y) = extendedMineSize' mine

loadMap :: FilePath -> IO (Meta,Mine)
loadMap fpath = do
  text <- readFile fpath
  let (meta,mine0) = parseMine text
      mine = extendMine mine0
  return (meta,mine)        

stdinMap :: IO (Meta,Mine)
stdinMap = do
  text <- hGetContents stdin
  let (meta,mine0) = parseMine text
      mine = extendMine mine0
  return (meta,mine)
            
--------------------------------------------------------------------------------
  
isLift :: Cell -> Bool
isLift (Lift _) = True
isLift _ = False

isRock :: Cell -> Bool
isRock (Rock _) = True
isRock _ = False

findRobot :: Mine -> Pos
findRobot mine = case find (\(xy,c) -> c==Robot) $ assocs mine of
  Nothing     -> error "findRobot: no robot!"
  Just (xy,_) -> xy

findLift :: Mine -> Pos
findLift mine = case find (\(xy,c) -> isLift c) $ assocs mine of
  Nothing     -> error "findLift: no lift!"
  Just (xy,_) -> xy

unsafeFindCell :: Cell -> Mine -> Pos
unsafeFindCell what mine = case find (\(xy,c) -> c==what) $ assocs mine of
  Nothing     -> error $ "unsafeFindCell: " ++ show what
  Just (xy,_) -> xy
  
findCells :: Cell -> Mine -> [Pos]
findCells what mine = [ xy | (xy,c) <- assocs mine , c==what ]

findExposedLambdas :: Mine -> [Pos]
findExposedLambdas = findCells Lambda

findHiddenLambdas :: Mine -> [Pos]
findHiddenLambdas = findCells (Rock True)

findAllLambdas :: Mine -> [Pos]
findAllLambdas m = findExposedLambdas m ++ findHiddenLambdas m

{-
-- atlosan bejarunk
-- azert az se gyors?
getNearCells :: Cell -> Int -> IOMine -> [Pos]
getNearCells what radius mine =
  do
    collector <- newIORef []
    wrapper collector
    readIORef collector
  where
    wrapper collector = go 0 0 pos where
      go radius phase pos = case phase of     
        0 -> go (radius+1) 1 (move pos R)
        1 -> do
          c <- readArray mine pos
          go  radius
-}

{-
xReadArray :: IOMine -> Pos -> Cell
xReadArray arr pos = getBounds arr >>= \bnd -> unsafeRead arr (Ix.index bnd pos)
-}
   
--------------------------------------------------------------------------------

-- all 8 neighbourhooding positions
neighbours :: Pos -> [Pos]
neighbours p = -- (Pos x y) = 
  [ mkPos (x-1)  y
  , mkPos (x-1) (y-1)
  , mkPos  x    (y-1)
  , mkPos (x+1) (y-1)
  , mkPos (x+1)  y 
  , mkPos (x+1) (y+1)
  , mkPos  x    (y+1)
  , mkPos (x-1) (y+1)
  ]
  where (x,y) = unPos p

ioFindBeards :: IOMine -> IO [Pos] 
ioFindBeards mine = do
  ass <- getAssocs mine
  return [ xy | (xy,Beard) <- ass ]

checkForBeards = checkForNeighbours Beard
checkForEmpty  = checkForNeighbours Empty
  
checkForNeighbours :: Cell -> IOMine -> Pos -> IO [Pos]
checkForNeighbours what mine pos = filterM worker (neighbours pos) where
  worker p = readArray mine p >>= \c -> return (c==what)  
  
--------------------------------------------------------------------------------

sum' :: Num a => [a] -> a
sum' = foldl' (+) 0

fmapMaybe :: (a -> b) -> Maybe a -> Maybe b
fmapMaybe = fmap

reverseOrdering :: Ordering -> Ordering
reverseOrdering LT = GT
reverseOrdering GT = LT
reverseOrdering EQ = EQ

reverseCompare :: Ord a => a -> a -> Ordering
reverseCompare x y = reverseOrdering $ compare x y

reverseComparing :: Ord b => (a -> b) -> a -> a -> Ordering
reverseComparing f x y = reverseOrdering $ compare (f x) (f y)

equating :: Eq b => (a -> b) -> a -> a -> Bool
equating f x y = (f x == f y)

--------------------------------------------------------------------------------

generateMoves :: Meta -> IOMine -> Pos -> IO [Move]
generateMoves meta mine pos = do
  -- mmms <- sequence [ canMove mine pos (Move dir) | dir <- allDirs ]     -- mmms = Maybe MoveMentS :)
  -- return ( Wait : [ mm | Just mm <- mmms ] ) 
  mds <- sequence [ canMoveDir meta mine pos dir | dir <- allDirs ] 
  return ( {- Wait : -} [ Move d | Just d <- mds ] ) 
  
canMoveDir :: Meta -> IOMine -> Pos -> Dir -> IO (Maybe Dir)
canMoveDir meta mine pos dir = do
  mmm <- canMove_noShave meta mine pos (Move dir)
  case mmm of 
    Just mm -> return (Just dir)
    Nothing -> return Nothing

--------------------------------------------------------------------------------

waitForSpace :: IO ()
waitForSpace = hSetBuffering stdin NoBuffering >> go where
  go = do
    c <- getChar
    case c of
      ' ' -> return ()
      '\ESC' -> do
        b2 <- hReady stdin
        if not b2 then go else do
          c2 <- getChar
          b3 <- hReady stdin
          if not b3 then go else do
            c3 <- getChar
            go
      _ -> go

-- Actually this has totally different target semantics, 
-- but we cheat and reuse existing data types 
-- (what we want is to forward/back + exit for movie playback)
getLeftRightAbort :: IO Move
getLeftRightAbort = go where
  go = do
    c <- getChar
    case toUpper c of
      'A' -> return Abort
      '\ESC' -> do
        b2 <- hReady stdin
        if not b2 then go else do
          c2 <- getChar
          b3 <- hReady stdin
          if not b3 then go else do
            c3 <- getChar
            case (ord c2,ord c3) of
              (91,68) -> return (Move L)
              (91,67) -> return (Move R)
              _ -> go
      _ -> go

getInteractiveMove :: IO Move
getInteractiveMove = go where
  go = do
    c <- getChar
    case toUpper c of
      'A' -> return Abort
      'W' -> return Wait
      'L' -> return (Move L)
      'D' -> return (Move D)
      'R' -> return (Move R)
      'U' -> return (Move U)
      'S' -> return Shave
      '\ESC' -> do
        b2 <- hReady stdin
        if not b2 then go else do
          c2 <- getChar
          b3 <- hReady stdin
          if not b3 then go else do
            c3 <- getChar
            case (ord c2,ord c3) of
              (91,68) -> return (Move L)        -- ANSI cursor input
              (91,67) -> return (Move R) 
              (91,65) -> return (Move U) 
              (91,66) -> return (Move D) 
              _ -> go
      _ -> go

testInteractiveMove :: IO ()
testInteractiveMove = 
  do
    hSetBuffering stdin NoBuffering
    hSetEcho stdin False
    go
  where
    go = do
      m <- getInteractiveMove
      print m
      when (m/=Abort) go

--------------------------------------------------------------------------------
  
      