
-- OK, 24h left, it's time to make a plan :)

{-# LANGUAGE BangPatterns #-}
module Planner where

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Maybe
import Data.Ord

import Control.Monad
import Control.Applicative
import Data.IORef

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import Ansi
import PosSet
import Spatial
import Mine
import Simulation
import SimBetter
import Search
import AStar

--------------------------------------------------------------------------------

data Target 
  = Nearby                -- ^ there is stuff very near, it makes sense to do DFS
  | Scan                  -- ^ just scan around
  | Faraway  !Pos !Cell   -- ^ explicit target
  | BeardTgt !Pos         -- ^ this is different because we want a neighbouring cell, not the beard itself
  deriving (Eq,Ord,Show)

data LocalPlan
  = Idle                 -- ^ do nothing
  | DepthFirst !Int      -- ^ brute force DFS with given depth
  | PrunedDFS  !Int !Int !(Maybe Int) -- ^ pruned tree of small DFS searches (depth, nsurvivors); we possibly limit the number of executed steps
  | Dijkstra   !Int      -- ^ Dijkstra in a given radius
  | AStarExact !Pos      -- ^ A* to the given destination
  | AStarFuzzy !Pos      -- ^ when we are near the target, we switch to DFS instead
  | Teleport   !Pos      -- ^ first we teleport with exact A*, then look around
  | Hutton     !Pos      -- ^ similar to AStarFuzzy but we shave at the end
  | JustShave            -- ^ "right here, right now"
  | DfsShave   !Int      -- ^ DFS + Shave at the end 
  | AbortPlan
  deriving (Eq,Ord,Show)
  
type Plan = [LocalPlan]  

--------------------------------------------------------------------------------

isDepthFirst :: LocalPlan -> Bool
isDepthFirst (DepthFirst _) = True
isDepthFirst _ = False

isPrunedDFS :: LocalPlan -> Bool
isPrunedDFS (PrunedDFS {}) = True
isPrunedDFS _ = False

notIdle :: LocalPlan -> Bool
notIdle Idle = False
notIdle _    = True

--------------------------------------------------------------------------------

{-
prunedPlanner :: SimState -> IO Plan
prunedPlanner = 
  do
    someplans <- prune 3 0 [[]]
    case someplans of 
      []       -> [DepthFirst 5]     -- nincs jobb otletunk :(
      allplans -> if bestscore >= currentscore - 4 then best else [AbortPlan] where
        (best,bestscore) = head $ sortBy (reverseComparing snd) allplans
  where

  prune 0 _       plans   = return plan 
  prune d fstiter oldplans = do
    m_allplans <- concat <$> forM oldplans $ \oldplan -> do
                    generalPlanner (3-fstiter) state (\plan state' -> prune (d-1) 1 return (plan, calculateStateScore state'))
                    
    return $ case m_allplans of 
      []       -> [DepthFirst 5]     -- nincs jobb otletunk :(
      allplans -> if bestscore >= currentscore - 3 then best else [AbortPlan] where
        (best,bestscore) = head $ sortBy (reverseComparing snd) allplans
  where
    currentscore = calculateStateScore state
-}

simplePlanner :: SimState -> IO Plan
simplePlanner state = do
  m_allplans <- generalPlanner 3 state (\plan moves state' -> return (filter notIdle plan, calculateStateScore state', moves ))

  whenDebug $ do
    printIOMine (_simMine state)
    print currentscore
    mapM_ (putStrLn . showTriple) m_allplans
  
  case m_allplans of 
    []       -> return [] -- AbortGame]   -- [DepthFirst 6]     -- nincs jobb otletunk :(
    allplans -> if bestscore >= currentscore - 4 
                  then do
                    whenDebug $ print $ "========\nbest plan is: " ++ showTriple triple
                    return best 
                  else do
                  
                    humanMessage "---- only bad scores => LAST CHANCE"
                    tryTheLastChance >>= \mstuff -> case mstuff of
                      Nothing -> do humanMessage "---- last chance failed, ABORT"
                                    return [AbortPlan]
                      Just (xplan,xscore,xmoves) -> if xscore >= currentscore - 9 
                        then return xplan
                        else do 
                          humanMessage "---- ABORT caused by having only bad scores (not even last chance helped)"
                          return [AbortPlan] 
                        
                where
                  triple@(best,bestscore,moves) = head $ sortBy (reverseComparing snd2) allplans
  where
    currentscore = calculateStateScore state
    snd2 (x,y,z) = y
    showTriple (best,score,moves) = show score ++ " | " ++ show best ++ " | " ++ show (map encodeMoves moves)

    tryTheLastChance = do
      let lplans = 
            [ PrunedDFS 12 200 Nothing
            , JustShave
            , DfsShave 3
            , DfsShave 6
            ]
      stuff <- forM lplans $ \lplan -> do
        simulateLocalPlan lplan state $ \moves state' -> do
          return ( [lplan], calculateStateScore state', moves )
      case sortBy (reverseComparing snd2) (catMaybes stuff) of
        []    -> return Nothing
        (p:_) -> return (Just p)
         
--------------------------------------------------------------------------------
    
-- we end any plan with a DFS (it tends to avoid suicide by pushing the prediction into the far future)
generalPlanner :: Int -> SimState -> (Plan -> [[Move]] -> SimState -> IO a) -> IO [a]
generalPlanner depth0 state0 action0 = 
  do
    collector <- newIORef []
    wrapper collector depth0 state0 action0
    readIORef collector
  where  
    startingscore = calculateStateScore state0
    scorelowerbound = startingscore - 100
    
    wrapper collector = go [] [] where
    
      go revplan revmoves 0 state action = do
        go revplan revmoves (-1) state action 
        {-
        let lplan = DepthFirst 3    -- last step is always this
        simulateLocalPlan lplan state $ \moves state' -> 
          go (lplan:revplan) (moves:revmoves) (-1) state' action
        return ()
        -}

      go revplan revmoves (-1) state action = do        
        x <- action (reverse revplan) (reverse revmoves) state 
        modifyIORef collector (x:)
        
      go revplan revmoves d state action = when (calculateStateScore state >= scorelowerbound) $ do

        targets <- selectTargets state         
        let lplans = concat $ flip map targets $ \target -> case target of
        
              Scan -> if d==depth0 
                then [ Dijkstra 20 , PrunedDFS 9 20 (Just 15)]  -- 125
                else case filter isPrunedDFS revplan of 
                  [] -> [ Idle , Dijkstra 12 , PrunedDFS 5 15 (Just 10)]      -- profiling szerint PrunedDFS viszi el az idot; masreszt tok hasznos 
                  _  -> [ Idle , Dijkstra 12 ]
                  
              Nearby -> [DepthFirst 5] 
                        -- case filter isDepthFirst revplan of { [] -> [ DepthFirst 5 ] ; _ -> [] } -- csak akkor ha meg nem volt
                        
              Faraway pos cell -> case cell of
                Lambda      -> [ AStarFuzzy pos ]
                Rock True   -> [ AStarFuzzy pos ]
                TrampSrc {} -> [ Teleport   pos ]
                _           -> [ AStarExact pos ]
                
              BeardTgt pos -> [ Hutton pos ]
                
        -- whenDebug $ putStrLn $ ">>> " ++ show targets ++ " => sel: " ++ show lplans ++ " ~~~ sofar: " ++ show (reverse revplan)
                        
        forM_ lplans $ \lplan -> 
          simulateLocalPlan lplan state $ \moves state' -> 
            go (lplan:revplan) (moves:revmoves) (d-1) state' action
       
--------------------------------------------------------------------------------

localPlanMoves :: LocalPlan -> SimState -> IO (Maybe [Move])
localPlanMoves lplan state = simulateLocalPlan lplan state (\moves state' -> return moves)

simulateLocalPlan :: LocalPlan -> SimState -> ([Move] -> SimState -> IO a) -> IO (Maybe a)
simulateLocalPlan localplan state cont = 
  case localplan of

    Idle -> Just <$> cont [] state

    JustShave -> do
      let moves = [Shave]
      simulateMovesWithUndo state moves $ \state' -> Just <$> cont moves state' 

    AbortPlan -> do
      let moves = [Abort]
      humanMessage "---- ABORT caused by AbortPlan"
      simulateMovesWithUndo state moves $ \state' -> Just <$> cont moves state' 
 
    DepthFirst depth  -> do
      moves <- {-# SCC "|DepthFirst|" #-} depthFirstSearchBest depth state
      simulateMovesWithUndo state moves $ \state' -> Just <$> cont moves state' 

    DfsShave depth  -> do
      moves <- {-# SCC "|DepthFirst/Shave|" #-} depthFirstSearchBest depth state
      simulateMovesWithUndo state moves $ \state' -> do 
        simulateMoveWithUndo state' Shave $ \state'' -> do   
          Just <$> cont (moves ++ [Shave]) state'' 
  
    Dijkstra radius -> do
      mpath <- {-# SCC "|Dijkstra|" #-} dijkstraBest radius state
      case mpath of
        Nothing   -> return Nothing
        Just path -> simulatePathWithUndo state path $ \state' -> Just <$> cont (pathMoves path) state' 
  
    PrunedDFS depth nsurvivors mexecute -> do
      moves0 <- {-# SCC "|PrunedDFS|" #-} prunedNextMoves depth nsurvivors  state    -- 10x3 = 30 steps
      let moves = case mexecute of
            Nothing -> moves0
            Just k  -> take k moves0
      simulateMovesWithUndo state moves $ \state' -> Just <$> cont moves state' 
      
    Teleport tgtpos -> do
      mpath <- {-# SCC "|AStar/Teleport|" #-} timeoutAStar defaultAStar state tgtpos
      case mpath of
        Nothing   -> return Nothing
        Just path -> simulatePathWithUndo state path $ \state' -> do
          moves' <- depthFirstSearchBest 6 state'
          -- moves' <- prunedNextMoves 10 100 state'
          simulateMovesWithUndo state' moves' $ \state'' -> Just <$> cont (pathMoves path ++ moves') state'' 
          
          -- mpath' <- {-# SCC "|Teleport/Dijkstra|" #-} dijkstraBest 13 state' 
          -- case mpath' of 
          --   Just path' -> simulatePathWithUndo state' path' $ \state'' -> Just <$> cont (pathMoves path ++ pathMoves path') state'' 
          --   Nothing -> return Nothing
            
    AStarExact tgtpos -> do
      mpath <- {-# SCC "|AStar/Exact|" #-} timeoutAStar defaultAStar state tgtpos
      case mpath of
        Nothing   -> return Nothing
        Just path -> simulatePathWithUndo state path $ \state' -> Just <$> cont (pathMoves path) state' 

    AStarFuzzy tgtpos -> do
      mpath <- {-# SCC "|Fuzzy/AStar|" #-} timeoutAStar defaultAStar state tgtpos
      case mpath of
        Nothing   -> return Nothing
        Just (Path path0) -> do
          let path1 = Path (drop 2 path0)                   -- last two step is omitted
          simulatePathWithUndo state path1 $ \state' -> do
            moves' <- {-# SCC "|Fuzzy/DFS|" #-} depthFirstSearchBest 5 state'         -- DFS for 5 steps
            simulateMovesWithUndo state' moves' $ \state'' -> Just <$> cont (pathMoves path1 ++ moves') state'' 

    Hutton tgtpos -> do
      mpath <- {-# SCC "|Hutton/AStar|" #-} timeoutAStar defaultAStar state tgtpos
      case mpath of
        Nothing   -> return Nothing
        Just (Path path0) -> do
          let path1 = Path (drop 3 path0)                   -- last three step is omitted
          simulatePathWithUndo state path1 $ \state' -> do
            moves' <- {-# SCC "|Hutton/DFS|" #-} depthFirstSearchBest 5 state'         -- DFS for 5 steps
            simulateMovesWithUndo state' moves' $ \state'' -> do
              simulateMoveWithUndo state'' Shave $ \state''' -> do   
                Just <$> cont (pathMoves path1 ++ moves' ++ [Shave]) state''' 

--------------------------------------------------------------------------------

isCellInteresting :: Cell -> Maybe Int
isCellInteresting c = case c of
  Lambda      -> Just 66
  Lift Open   -> Just 10000
  Razor       -> Just 300
  TrampSrc {} -> Just 100
  _           -> Nothing
  
--------------------------------------------------------------------------------

selectTargets :: SimState -> IO [Target]
selectTargets state = 
  do  
    btargets <- beardTargets
    whenDebug $ print (ourpos,btargets,alltargets)
    targets <- forM (alltargets) $ \p -> do
      c <- readArray (_simMine state) p
      return (Faraway p c)      
    return $ Scan : (if null near then targets else (Nearby : targets)) ++ btargets
    
  where
    ourpos  = _simRobotPos  state
    lambdas = _simLambdaSet state
    beards  = _simBeardSet  state
    
    teleports = Map.keys (_simTrampolins state)
    
    alltargets = teltargets ++ take 3 lamtargets
    
    lamtargets = take (4 - length teltargets) lamtargets_all 
    teltargets = take 2 (sortBy (comparing distFromUs) teleports)
    
    distance = astarDist
    distFromUs !p = distance ourpos p 

--    sortedlambdas = sortBy (comparing distFromUs) (  bsetToList lambdas) where
    sortedlambdas = 
      case stateCountRemainingLambdas state of
        0 -> []
        _ -> sortBy (comparing distFromUs) $ nonemptyBSetFindClosest 10 500 ourpos lambdas

    (near,far) = span (\p -> distFromUs p <= 5) sortedlambdas
    
    beardTargets :: IO [Target]
    beardTargets = if (_simBeardCnt state <= 0)
      then return []
      else do
        let bs = bsetFindClose 10 ourpos beards
        case sortBy (comparing distFromUs) bs of
          [] -> return [] 
          (beard:_) -> do
            -- now, a beard is not reachable, so let's find a free cell around
            let ps = neighbours beard
                mine = _simMine state
            ns <- sequence [ readArray mine p >>= \c -> return (p,c) | p <- ps ] 
            case sortBy (comparing distFromUs) [ p | (p,c) <- ns , cellIsReachable c ] of
               []             -> return []     
               (nowthisone:_) -> return [BeardTgt nowthisone]
               
    -- We select targets in the following order: 
    -- we chose the next target to be the furthest from all the existing targets
    -- but still not too far from us ?     
    lamtargets_all = 
      case far of
        []     -> [_simLift state]
        (p:ps) -> go [p] ps
      where
        go ts [] = ts
        go ts ps = let (q:qs) = sortBy (comparing f) ps in go (q:ts) qs where
          f p = sum' [ d - 0.33 * distance p t | t<-ts ] where d = distance p ourpos
            
--------------------------------------------------------------------------------
        
plannerStrategy :: Meta -> Mine -> IO [Move]    
plannerStrategy meta puremine = 
  do
    inistate  <- initialSimState meta puremine
    collector <- newIORef [] :: IO (IORef [Move]) 
    wrapper collector inistate
    reverse <$> readIORef collector
  where
    maxStepsAllowed = extendedMineSize puremine - 1
    wrapper collector inistate = go [] inistate where
    
      go (this:rest) state = makeMove this state (go rest)
      go [] state = do
        moves <- if ( _simMoveCnt state > maxStepsAllowed )
          then do
            humanMessage "---- ABORT caused by empty move list"
            return [Abort]
          else do
            plan <- simplePlanner state
            case plan of
              [] -> do
                humanMessage "---- ABORT caused by empty plan"
                return [Abort]
              (lplan:_) -> do
                mmoves <- localPlanMoves lplan state 
                case mmoves of 
                  Nothing    -> return [Wait]
                  Just moves -> return moves
                
        go moves state

      makeMove move state cont = do
        modifyIORef collector (move:)
        humanPrint move
        whenSubmission $ putChar (encodeMove move)
        st <- simulateStep state move  
        case eiSimState st of
          Right state' -> cont state'
          Left  (result,state') -> humanPrint result         
           
--------------------------------------------------------------------------------
      
the_strategy = plannerStrategy

        