

-- | Generic (not algorithm-specific) part of game simulation 

module Simulation where 

--------------------------------------------------------------------------------

import Data.Array
import Data.Array.IO
import Data.Array.MArray

import Data.List
import Data.Maybe

import Control.Monad
import Control.Applicative

import Data.IORef
import System.IO

import Data.Map (Map) ; import qualified Data.Map as Map
import Data.Set (Set) ; import qualified Data.Set as Set

import PosSet
import Spatial
import Mine
import Ansi

--------------------------------------------------------------------------------

{-
data Status = Status
  { _statScore :: !Double
  }
  deriving (Eq,Ord,Show)
-}
type Status = Double
_statScore :: Status -> Double
_statScore = id

-------------   
    
-- | One step in a long path towards enlightenment...
data Step = Step 
  { _stepMove   :: !Move 
  , _stepStatus :: !Status 
  , _stepChange :: !ChangeSet 
  }
  deriving (Eq,Ord,Show)
  
-- | The Royal Path, but in /REVERSE/ order!!!
newtype Path = Path [Step] deriving (Eq,Ord) -- ,Show)

instance Show Path where show path@(Path steps) = "<" ++ encodeMoves (pathMoves path) ++ ">"

pathMoves :: Path -> [Move]
pathMoves (Path steps) = reverse $ map _stepMove steps

emptyPath :: Path
emptyPath = Path []

pathStatus :: Path -> Status
pathStatus (Path steps) = _stepStatus (head steps)

pathScore :: Path -> Double
pathScore (Path steps) = case steps of
  [] -> 0
  (s:ss) -> _statScore (_stepStatus s)

--------------------------------------------------------------------------------

undoChangeSet :: IOMine -> ChangeSet -> IO ()
undoChangeSet mine (ChangeSet writes undos) = executeUndo mine undos 

{- 
-- we don't need such arcane stuff like non-singleton state :)
cloneState :: SimState -> IO SimState
cloneState simstate = do
  frozen <- unsafeFreeze (_simMine simstate) :: IO Mine
  mine'  <- thaw frozen
  return $ simstate { _simMine = mine' }
-}

--------------------------------------------------------------------------------

simulateMoveWithUndo' :: SimulationStep -> SimState -> Move -> (SimState -> IO a) -> IO a
simulateMoveWithUndo' simulateStep state move action = do
  let themine = _simMine state 
  state' <- simulateStep state move
  let changes = _simLastChanges state'
  x <- action state'
  undoChangeSet themine changes
  return x

simulateMovesWithUndo' :: SimulationStep -> SimState -> [Move] -> (SimState -> IO a) -> IO a
simulateMovesWithUndo' simulateStep state0 moves action = go moves state0 where
  go []           state = action state
  go (move:moves) state = if stateIsFinished state 
    then action state
    else simulateMoveWithUndo' simulateStep state move (go moves) 

simulatePathWithUndo' :: SimulationStep -> SimState -> Path -> (SimState -> IO a) -> IO a
simulatePathWithUndo' simulateStep state0 (Path path) action = go (reverse path) state0 where
  go []           state = action state
  go (step:steps) state = if stateIsFinished state 
    then action state
    else do
      let move = _stepMove step
      simulateMoveWithUndo' simulateStep state move (go steps) 

--------------------------------------------------------------------------------
       
type EiSimResult = Either (Result,SimState) SimState
       
type SimulationStep = SimState -> Move -> IO SimState -- EiSimResult

--------------------------------------------------------------------------------
    
data SimState = SimState 
  { _simRobotPos     :: !Pos 
  , _simLift         :: !Pos
  , _simLambdaCnt    :: !Int         -- ^ lambdas collected
  , _simLambdaSet    :: !(BSet)      -- ^ remaining lambdas 
  , _simBeardSet     :: !(BSet)      -- ^ position of beards
  , _simTrampolins   :: !(Map Pos Pos)  -- ^ trampolines (from,to)
  , _simHorockSet    :: !(BSet)      -- ^ set of higher-order rocks
  , _simTotalLambda  :: !Int         -- ^ total lambda
  , _simTotLamNoRock  :: !Int         -- ^ total lambda
  , _simMoveCnt      :: !Int         -- ^ movement = step (according to web validator, WWWWA = -5 points)
  , _simRazors       :: !Int         -- ^ number of razors available
  , _simBeardCnt     :: !Int         -- ^ beard population count 
  , _simWaterLevel   :: !Int         -- ^ current water level
  , _simUnderWater   :: !Int         -- ^ how long are we underwater
  , _simRevHistory   :: [(Pos,Move)] -- ^ history of moves (from,move) in reverse order
  , _simWatch        :: !Watch       -- ^ positions to watch at the next update step
  , _simLastChanges  :: !ChangeSet   -- ^ the map changes done in the last space
  , _simLastMovement :: !(Maybe Movement)   -- ^ last movement
  , _simSumMMScore   :: !Double             -- ^ sum of movement scores
  , _simResult       :: !(Maybe Result)   -- ^ in the case we reached an ending condition... better unify data
  , _simMine         :: !IOMine 
  , _simMeta         :: !Meta
  }
  
stateCountRemainingLambdas :: SimState -> Int
stateCountRemainingLambdas state = (_simTotLamNoRock state - _simLambdaCnt state)
  
stateIsFinished :: SimState -> Bool
stateIsFinished state = case _simResult state of 
  Nothing -> False
  Just {} -> True

stateResult :: SimState -> Maybe Result
stateResult = _simResult

stateEndCond :: SimState -> Maybe EndCond
stateEndCond s = case stateResult s of
  Just r  -> Just (_endCond r)
  Nothing -> Nothing
     
-- backward compatibility
eiSimState :: SimState -> EiSimResult
eiSimState state = case stateIsFinished state of 
  False -> Right state
  True  -> Left (fromJust (_simResult state), state)
  
--------------------------------------------------------------------------------

initialSimState :: Meta -> Mine -> IO SimState 
initialSimState meta initial = 
  do

    mine <- thaw initial
    
    let flooding = _metaFlooding meta
        hair     = _metaHair     meta
        
        iniState = SimState 
           { _simRobotPos    = inirobotpos 
           , _simLift        = liftpos
           , _simLambdaCnt   = 0
           , _simTotalLambda = totalLamCnt
           , _simTotLamNoRock = length expLambdas
           , _simLambdaSet   = bsetFromList expLambdas
           , _simBeardSet    = bsetFromList allBeards
           , _simTrampolins  = Map.fromList tramps
           , _simHorockSet   = bsetFromList horocks
           , _simMoveCnt     = 0
           , _simRazors      = _initialRazors hair
           , _simBeardCnt    = length allBeards
           , _simWaterLevel  = _waterLevel flooding
           , _simUnderWater  = 0
           , _simRevHistory  = []
           , _simWatch       = mineWatch initial
           , _simLastChanges = ChangeSet [] []
           , _simLastMovement = Nothing
           , _simSumMMScore   = 0
           , _simResult      = Nothing
           , _simMine        = mine 
           , _simMeta        = meta
           }

    return iniState
    
  where
  
    inirobotpos = findRobot initial
    liftpos     = findLift  initial
    expLambdas  = findExposedLambdas initial
    allBeards   = findCells Beard initial
    totalLamCnt = length (findAllLambdas initial)
    horocks     = findCells (Rock True) initial

    tramps = [ (src,tgt) | Trampoline (src,_) (tgt,_) <- _metaTrampolines meta ]

--------------------------------------------------------------------------------

canMove :: SimState -> Pos -> Move -> IO (Maybe Movement)
canMove state pos mv =
  case mv of
    Abort -> return $ Just $ AbortGame pos
    Wait  -> return $ Just $ NoMove    pos 
    Move dir -> canMove' meta mine pos dir
    Shave -> if nrazors>0 
      then do
        what <- checkForBeards mine pos
        return $ Just (Shaved pos what) 
      else return $ Just (NoRazor pos)
  where
    meta = _simMeta state
    mine = _simMine state
    nrazors = _simRazors state

--------------------------------------------------------------------------------

printStateInfo :: SimState -> IO ()
printStateInfo oldState = do      
  putStrLn "\n"
  putStrLn $ concat
    [ "lambda = " , show (_simLambdaCnt oldState) , " / " , show (_simTotalLambda oldState)
    , "  |  moves = " , show (_simMoveCnt    oldState)  
    , "  |  ~water = " , show (_simWaterLevel oldState) , " (" , show (_simUnderWater oldState) , ")"
    , "  |  robot = " , show (_simRobotPos   oldState)
    , "  |  #razors = " , show (_simRazors  oldState) 
    ]

dumpStateInfo :: SimState -> IO ()
dumpStateInfo oldState = do      
  putStrLn""
  putStrLn $ concat
    [ "lambda=" , show (_simLambdaCnt oldState) , "/" , show (_simTotalLambda oldState)
    , ", moves=" , show (_simMoveCnt    oldState)  
    , ", water=" , show (_simWaterLevel oldState) , " (" , show (_simUnderWater oldState) , ")"
    , ", razor=" , show (_simRazors  oldState) 
    ]

--------------------------------------------------------------------------------

interactiveSimWith :: SimulationStep -> Meta -> Mine -> IO ()
interactiveSimWith simStep meta initial =
  do
    writeIORef submissionMode False
          
    hSetBuffering stdin NoBuffering
    hSetEcho stdin False
          
    iniState <- initialSimState meta initial 
    
    go iniState
    
  where
    
    go oldState = do
      clrscr >> topleft
      printStateInfo oldState
      printIOMineWater (_simWaterLevel oldState) (_simMine oldState)
      
      -- print $ (_simLambdaSet oldState)
      -- print $ (_simHorockSet oldState)
      -- print $ (_simTrampolins oldState) 
      -- print $ (_simBeardCnt oldState)
      -- print $ (bsetToUnsortedList $ _simBeardSet oldState)
      
      move <- getInteractiveMove
      print move
      st <- simStep oldState move 
      case eiSimState st of
        Left (result,newstate) -> do
          putStrLn "" >> print result
          saveHistory "_last.hist" (_history result)
          putStrLn $ "total points = " ++ show (resultPoints result)
        Right newState -> go newState 

--------------------------------------------------------------------------------

-- | We just maje a big ascii dump, so we can compare different implementations
dumpSimulationWith :: SimulationStep -> Meta -> Mine -> [Move] -> IO ()
dumpSimulationWith simStep meta initial moves =
  do 
    iniState <- initialSimState meta initial 
    go moves iniState
     
  where
    go [] oldState = do
      -- utolso allapotot is ki kell irni ha nem nyertunk/vesztettunk
      dumpStateInfo oldState
      dumpExtendedIOMineAscii (_simMine oldState)
    
    go (move:moves) oldState = do
      dumpStateInfo oldState
      dumpExtendedIOMineAscii (_simMine oldState)
      putStrLn [encodeMove move]
      simulateMoveWithUndo' simStep oldState move $ \state_ -> case eiSimState state_ of
        Left (result,newState) -> do
          dumpStateInfo newState
          dumpExtendedIOMineAscii (_simMine newState)
          
          putStrLn $ case _endCond result of
            Winning -> "WIN"
            Aborted -> "ABORT"
            Losing  -> "DIE(rock)"
            Inoperative -> "DIE(water)"
          putStrLn $ "total points = " ++ show (resultPoints result)
          return ()
        Right newState -> do
          go moves newState 

--------------------------------------------------------------------------------

-- list zipper 
data Zipper a = Zipper [a] [a]

front :: [a] -> Zipper a
front xs = Zipper [] xs

rfocus :: Zipper a -> Maybe a
rfocus (Zipper _ (r:rs)) = Just r
rfocus (Zipper _ []    ) = Nothing

zipperToList :: Zipper a -> [a]
zipperToList (Zipper ls rs) = reverse ls ++ rs

moveRight :: Zipper a -> Maybe (Zipper a)
moveRight (Zipper ls (r:rs)) = Just $ Zipper (r:ls) rs
moveRight (Zipper ls []    ) = Nothing

moveLeft :: Zipper a -> Maybe (Zipper a)
moveLeft (Zipper (l:ls) rs) = Just $ Zipper ls (l:rs)
moveLeft (Zipper []     rs) = Nothing

--------------------------------------------------------------------------------

zipperSimulationWith :: SimulationStep -> Meta -> Mine -> [Move] -> IO (Zipper (Move,SimState))
zipperSimulationWith simStep meta initial moves =
  do 
    iniState <- initialSimState meta initial 
    list <- go moves iniState
    return $ front list 
  where
    go [] oldState = return [(Wait,oldState)]
    go (move:moves) oldState = do
      print move
      simulateMoveWithUndo' simStep oldState move $ \state_ -> do
  --      ei <- simStep oldState move 
        case eiSimState state_ of
          Left (result,newState) -> do
            putStrLn "" >> print result
            putStrLn $ "total points = " ++ show (resultPoints result)
            return [(move,oldState)]
          Right newState -> do
            stuff <- go moves newState 
            return ( (move,oldState) : stuff )

zipperPlaybackWith :: SimulationStep -> Meta -> Mine -> Zipper (Move,SimState) -> IO ()
zipperPlaybackWith simStep meta initial zipper =
  do
    writeIORef submissionMode False
          
    hSetBuffering stdin NoBuffering
    hSetEcho stdin False
          
--    iniState <- initialSimState meta initial 
    
    let mine = snd $ fromJust $ rfocus zipper 
    go zipper mine -- iniState
    
  where
    
    go zipper oldState = do
      let themine = _simMine oldState
      clrscr >> topleft
      printStateInfo oldState
      printIOMineWater (_simWaterLevel oldState) (_simMine oldState)
      -- print $ (_simLambdaSet oldState)

      when (stateIsFinished oldState) $ do
        let result = fromJust $ _simResult oldState
        putStrLn "" >> print (_endCond result)
        putStrLn $ "total points = " ++ show (resultPoints result)
       
      leftright <- getLeftRightAbort
      case leftright of
     
        Abort -> putStrLn "\nexiting...\n"
     
        Move L -> case moveLeft zipper of
          Nothing -> go zipper oldState
          Just zipper' -> do
            let changes = _simLastChanges oldState
            -- print changes
            let (_,newState) = fromJust $ rfocus zipper'
            undoChangeSet (_simMine newState) changes
            go zipper' newState
                 
        Move R -> case moveRight zipper of
          Nothing -> go zipper oldState
          Just zipper' -> do
            let (move,_) = fromJust $ rfocus zipper
            print move
            state_ <- simStep oldState move 
            go zipper' state_

--------------------------------------------------------------------------------

